@@ -13,6 +13,8 @@
 #import <AVFoundation/AVFoundation.h>
 #import <CommonCrypto/CommonDigest.h>
 #import "UIDevice+SpeedCategory.h"
+#import "VLCBugreporter.h"
+#import <MediaPlayer/MediaPlayer.h>
 
 #define INPUT_RATE_DEFAULT  1000.
 
@@ -37,6 +39,7 @@
     BOOL _viewAppeared;
     BOOL _displayRemainingTime;
     BOOL _positionSet;
+    BOOL _playerIsSetup;
 }
 
 @property (nonatomic, strong) UIPopoverController *masterPopoverController;
@@ -61,13 +64,27 @@
 
 - (void)setMediaItem:(id)newMediaItem
 {
-    if (_mediaItem != newMediaItem)
+    if (_mediaItem != newMediaItem) {
+        [self _stopPlayback];
         _mediaItem = newMediaItem;
+        if (_viewAppeared)
+            [self _startPlayback];
+    }
 
     if (self.masterPopoverController != nil)
         [self.masterPopoverController dismissPopoverAnimated:YES];
 }
 
+- (void)setUrl:(NSURL *)url
+{
+    if (_url != url) {
+        [self _stopPlayback];
+        _url = url;
+        if (_viewAppeared)
+            [self _startPlayback];
+    }
+}
+
 - (void)viewDidLoad
 {
     [super viewDidLoad];
@@ -162,6 +179,11 @@
 
     if (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
         self.positionSlider.scrubbingSpeedChangePositions = @[@(0.), @(100.), @(200.), @(300)];
+
+    _playerIsSetup = NO;
+
+    [self.movieView setAccessibilityLabel:NSLocalizedString(@"VO_VIDEOPLAYER_TITLE", @"")];
+    [self.movieView setAccessibilityHint:NSLocalizedString(@"VO_VIDEOPLAYER_DOUBLETAP", @"")];
 }
 
 - (BOOL)_blobCheck
@@ -189,16 +211,32 @@
 
 - (void)viewWillAppear:(BOOL)animated
 {
+    [super viewWillAppear:animated];
+
+    [self.navigationController setNavigationBarHidden:YES animated:YES];
+
+    if (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPhone)
+        [UIApplication sharedApplication].statusBarStyle = UIStatusBarStyleBlackTranslucent;
+
+    [self _startPlayback];
+
+    [self setControlsHidden:NO animated:YES];
+    _viewAppeared = YES;
+}
+
+- (void)_startPlayback
+{
+    if (_playerIsSetup)
+        return;
+
     _mediaPlayer = [[VLCMediaPlayer alloc] init];
     [_mediaPlayer setDelegate:self];
     [_mediaPlayer setDrawable:self.movieView];
 
-    [self.navigationController setNavigationBarHidden:YES animated:YES];
-    if (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPhone)
-        [UIApplication sharedApplication].statusBarStyle = UIStatusBarStyleBlackTranslucent;
-
-    if (!self.mediaItem && !self.url)
+    if (!self.mediaItem && !self.url) {
+        [self _stopPlayback];
         return;
+    }
 
     VLCMedia *media;
     if (self.mediaItem) {
@@ -213,7 +251,7 @@
     NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
     [media addOptions:
      @{kVLCSettingStretchAudio :
-           [[defaults objectForKey:kVLCSettingStretchAudio] boolValue] ? kVLCSettingStretchAudioOnValue : kVLCSettingStretchAudioOffValue, kVLCSettingTextEncoding : [defaults objectForKey:kVLCSettingTextEncoding], kVLCSettingSkipLoopFilter : [defaults objectForKey:kVLCSettingSkipLoopFilter]}];
+     [[defaults objectForKey:kVLCSettingStretchAudio] boolValue] ? kVLCSettingStretchAudioOnValue : kVLCSettingStretchAudioOffValue, kVLCSettingTextEncoding : [defaults objectForKey:kVLCSettingTextEncoding], kVLCSettingSkipLoopFilter : [defaults objectForKey:kVLCSettingSkipLoopFilter]}];
 
     [NSTimeZone resetSystemTimeZone];
     NSString *tzName = [[NSTimeZone systemTimeZone] name];
@@ -261,8 +299,6 @@
     self.positionSlider.value = 0.;
     [self.timeDisplay setTitle:@"" forState:UIControlStateNormal];
 
-    [super viewWillAppear:animated];
-
     if (![self _isMediaSuitableForDevice]) {
         UIAlertView * alert = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"DEVICE_TOOSLOW_TITLE", @"") message:[NSString stringWithFormat:NSLocalizedString(@"DEVICE_TOOSLOW", @""), [[UIDevice currentDevice] model], self.mediaItem.title] delegate:self cancelButtonTitle:NSLocalizedString(@"BUTTON_CANCEL", @"") otherButtonTitles:NSLocalizedString(@"BUTTON_OPEN", @""), nil];
         [alert show];
@@ -271,9 +307,6 @@
 
     if (![self hasExternalDisplay])
         self.brightnessSlider.value = [UIScreen mainScreen].brightness * 2.;
-
-    [self setControlsHidden:NO animated:YES];
-    _viewAppeared = YES;
 }
 
 - (BOOL)_isMediaSuitableForDevice
@@ -303,8 +336,10 @@
 {
     if (buttonIndex == 1)
         [self _playNewMedia];
-    else
+    else {
+        [self _stopPlayback];
         [self closePlayback:nil];
+    }
 }
 
 - (void)_playNewMedia
@@ -319,17 +354,28 @@
 
     [_mediaPlayer play];
 
+    if (self.mediaItem) {
+        if (self.mediaItem.lastAudioTrack.intValue > 0)
+            _mediaPlayer.currentAudioTrackIndex = self.mediaItem.lastAudioTrack.intValue;
+        if (self.mediaItem.lastSubtitleTrack.intValue > 0)
+            _mediaPlayer.currentVideoSubTitleIndex = self.mediaItem.lastSubtitleTrack.intValue;
+    }
+
     self.playbackSpeedSlider.value = [self _playbackSpeed];
     [self _updatePlaybackSpeedIndicator];
 
+    [self performSelectorInBackground:@selector(_updateExportedPlaybackInformation) withObject:nil];
+
     _currentAspectRatioMask = 0;
     _mediaPlayer.videoAspectRatio =  NULL;
 
     [self _resetIdleTimer];
+    _playerIsSetup = YES;
 }
 
 - (void)viewWillDisappear:(BOOL)animated
 {
+    [self _stopPlayback];
     _viewAppeared = NO;
     if (_idleTimer) {
         [_idleTimer invalidate];
@@ -338,12 +384,7 @@
     [self.navigationController setNavigationBarHidden:NO animated:YES];
     [UIApplication sharedApplication].statusBarStyle = UIStatusBarStyleBlackOpaque;
     [[UIApplication sharedApplication] setStatusBarHidden:NO withAnimation:UIStatusBarAnimationFade];
-    [_mediaPlayer pause];
     [super viewWillDisappear:animated];
-    if (self.mediaItem)
-        self.mediaItem.lastPosition = @([_mediaPlayer position]);
-    [_mediaPlayer stop];
-    _mediaPlayer = nil; // save memory and some CPU time
 
     // hide filter UI for next run
     if (!_videoFiltersHidden)
@@ -353,12 +394,27 @@
         _playbackSpeedViewHidden = YES;
 }
 
-- (id)initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil
+- (void)_stopPlayback
 {
-    self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil];
-    if (self)
-        self.title = @"Video Playback";
-    return self;
+    if (_mediaPlayer) {
+        [_mediaPlayer pause];
+        [self _saveCurrentState];
+        [_mediaPlayer stop];
+        _mediaPlayer = nil; // save memory and some CPU time
+    }
+    if (_mediaItem)
+        _mediaItem = nil;
+
+    _playerIsSetup = NO;
+}
+
+- (void)_saveCurrentState
+{
+    if (self.mediaItem) {
+        self.mediaItem.lastPosition = @([_mediaPlayer position]);
+        self.mediaItem.lastAudioTrack = @(_mediaPlayer.currentAudioTrackIndex);
+        self.mediaItem.lastSubtitleTrack = @(_mediaPlayer.currentVideoSubTitleIndex);
+    }
 }
 
 #pragma mark - remote events
@@ -384,6 +440,13 @@
     return YES;
 }
 
+
+- (void)motionEnded:(UIEventSubtype)motion withEvent:(UIEvent *)event
+{
+    if (motion == UIEventSubtypeMotionShake)
+        [[VLCBugreporter sharedInstance] handleBugreportRequest];
+}
+
 - (void)remoteControlReceivedWithEvent:(UIEvent *)event
 {
     switch (event.subtype) {
@@ -771,6 +834,9 @@
     float f_value = self.playbackSpeedSlider.value;
     double speed =  pow(2, f_value / 17.);
     self.playbackSpeedIndicator.text = [NSString stringWithFormat:@"%.2fx", speed];
+
+    /* rate changed, so update the exported info */
+    [self performSelectorInBackground:@selector(_updateExportedPlaybackInformation) withObject:nil];
 }
 
 - (float)_playbackSpeed
@@ -844,8 +910,7 @@
 
 - (void)applicationWillResignActive:(NSNotification *)aNotification
 {
-    if (self.mediaItem)
-        self.mediaItem.lastPosition = @([_mediaPlayer position]);
+    [self _saveCurrentState];
 
     if (![[[NSUserDefaults standardUserDefaults] objectForKey:kVLCSettingContinueAudioInBackgroundKey] boolValue]) {
         [_mediaPlayer pause];
@@ -868,6 +933,24 @@
         _mediaPlayer.currentVideoTrackIndex = 1;
 }
 
+- (void)_updateExportedPlaybackInformation
+{
+    if (!_mediaItem) {
+        [MPNowPlayingInfoCenter defaultCenter].nowPlayingInfo = nil;
+        return;
+    }
+    MLFile * currentFile = _mediaItem;
+
+    /* we omit artwork for now since we had to read it from storage as we can't access
+     * the artwork cache at the moment - FIXME? */
+    NSDictionary *currentlyPlayingTrackInfo = @{ MPMediaItemPropertyTitle : currentFile.title,
+                                                 MPMediaItemPropertyPlaybackDuration : @(currentFile.duration.intValue / 1000.),
+                                                 MPNowPlayingInfoPropertyElapsedPlaybackTime : @(_mediaPlayer.time.intValue / 1000.),
+                                                 MPNowPlayingInfoPropertyPlaybackRate : @(_mediaPlayer.rate) };
+
+    [MPNowPlayingInfoCenter defaultCenter].nowPlayingInfo = currentlyPlayingTrackInfo;
+}
+
 #pragma mark - autorotation
 
 - (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)toInterfaceOrientation {
